/*********************************************************************
 * File: at89s8252.c
 *
 * Program for reading and writing code (flash) and data (eeprom) 
 * memory space of the Atmel AT89S8252 via the Atmel STK200 programming
 * interface 
 */
#define VERSION		"0.1"
/*
 ************************************************************************/
#include <stdio.h>
#include <string.h>
#include <unistd.h>             /* getopt(), write(), close() */
#include <sys/types.h>          /* open() */
#include <sys/stat.h>           /* open() */
#include <fcntl.h>              /* open() */
#include <stdlib.h>		/* strtoul(), malloc() */
#include <string.h>
#include <ctype.h>		/* atoi() */                     

/*-------------------- local headers ---------------------------------------*/
#ifdef extern
#  undef extern
#endif
#define extern
#  include "at89s8252.h"
#  include "debug.h"
#undef extern

#include "port-io.h"
#include "loaddata.h"

/*-------------------- defines ---------------------------------------------*/
#define EXT_DELIM	'.'
#define PAR_DELIM	','

/*-------------------- variables -------------------------------------------*/
static PORT_T Ports[] = {
		{ "lp0",	0x378,	},
		{ "lp1",	0x278,	},
		{ "lp2",	0x3bc,	},
		/*--------- End ------------*/
		{ NULL,		0,	}, };

static LOCKBIT_T *At89s8252LockBits[] = { &Lb1, &Lb2, &Lb3, NULL };

static DEVICE_T Devices[] = {
		{ "AT89S8252",	8*1024, 2*1024,  At89s8252LockBits,	},
		/*--------- End ------------------------------------------*/
		{ NULL,		0,	0,	NULL,			} };

static char *BinExtension[] = { 
                "bin", 
                "rom", 
                "img",
                /*------*/
                NULL    };
static char *SRecExtension[] = {
                "srec",
                "mot",
                "moto",
                "rec",
                /*-----*/
                NULL    };
static char *IHexExtension[] = {
                "ihex",
                "hex",
                "ihx",
                "int",
                "intel",
                /*-----*/
                NULL    };      
 	
/*-------------------- Command line parameters --------------------------*/
#define PAR_CODE_BIN_FMT	'b'
#define PAR_DATA_BIN_FMT	'B'

#define PAR_CODE_IN_FILE	'c'
#define PAR_CODE_OUT_FILE	'C'

#define PAR_DATA_IN_FILE	'd'
#define PAR_DATA_OUT_FILE	'D'

#define PAR_ERASE		'e'

#define PAR_CODE_IHEX_FMT	'i'
#define PAR_DATA_IHEX_FMT 	'I'

#define PAR_LOCKBIT		'l'

#define PAR_DEBUG_LVL		'L'

#define PAR_PROG		'p'
#define PAR_PORT		'P'

#define PAR_CODE_SREC_FMT	's'
#define PAR_DATA_SREC_FMT	'S'

#define PAR_DEVICE		't'

#define PAR_VERIFY		'v'

static char CmdLine[] = { 	PAR_CODE_IN_FILE,	':',
				PAR_CODE_OUT_FILE,	':',
				PAR_DATA_IN_FILE,	':',
				PAR_DATA_OUT_FILE,	':',
				PAR_ERASE,
				PAR_LOCKBIT,		':',
				PAR_DEBUG_LVL,		':',
				PAR_PORT,		':',
				PAR_DEVICE,		':',
				PAR_VERIFY,
				PAR_PROG,
				PAR_CODE_BIN_FMT,
				PAR_DATA_BIN_FMT,
				PAR_DATA_IHEX_FMT,
				PAR_CODE_IHEX_FMT,
				PAR_DATA_SREC_FMT,
				PAR_CODE_SREC_FMT,
				/*------------- end --------------------*/
				'\0'					};

/*----------------------------- variables ----------------------------*/
static char		*PrgName;
static PORT_T		*Port		= Ports+0;
static DEVICE_T		*Device		= Devices+0;
static FILE		*DataIn 	= NULL;
static FILE		*CodeIn 	= NULL;
static int		DataOut		= -1;
static int		CodeOut		= -1;
static DATA_T		DataFmt		= UNDEF;
static DATA_T		CodeFmt		= UNDEF;
static DATA_T		SuggestDataFmt	= UNDEF;
static DATA_T		SuggestCodeFmt	= UNDEF;
static int		Verify		= 0;
static int		Programm	= 0;
static int		Erase		= 0;
static LOCKBIT_T	*Lock[sizeof(LockBits)/sizeof(LockBits[0])] = { NULL, };

/*----------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
static int ShowDeviceIds(FILE *Fp, int Num, int Offs) {
  int i;

  for(i = Offs; Devices[i].Id != NULL; i++) {
    fprintf(Fp,"%s",Devices[i].Id);
    if(Devices[i+1].Id == NULL) {
      fprintf(Fp,"\n");
      return(0);
    } else {
      if(i == Offs+Num-1) { 
        fprintf(Fp,"\n");
        return(Num + Offs);
      } else {
        fprintf(Fp,",");
      }
    } /* if(Devices[i+1].Id == NULL) */
  } /* for(i = Offs; Devices[i].Id != NULL; i++) */
  return(0);
}

/*----------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
static int ShowPortIds(FILE *Fp, int Num, int Offs) {
  int i;

  for(i = Offs; Ports[i].Id != NULL; i++) {
    fprintf(Fp,"%s",Ports[i].Id);
    if(Ports[i+1].Id == NULL) {
      fprintf(Fp,"\n");
      return(0);
    } else {
      if(i == Offs+Num-1) { 
        fprintf(Fp,"\n");
        return(Num + Offs);
      } else {
        fprintf(Fp,",");
      } 
    } /* else -- if(Ports[i+1].Id == NULL) */ 
  } /* for(i = Offs; Ports[i].Id != NULL; i++) */
  return(0);
}

/*----------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
static void ShowLockBits(FILE *Fp) {
  int i;
  for(i = 0; LockBits[i] != NULL; i++) { 
    fprintf(Fp, "%s", LockBits[i]->Id);
    if(LockBits[i+1]!=NULL) fprintf(Fp, ",");
  }
}

/*----------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
static void Usage(void) {
  int Offset = 0;

  fprintf(stderr,  "%s: Ver:" VERSION "(Built:" __DATE__ "/" __TIME__ ")\n",PrgName);
  fprintf(stderr,  "Usage: %s <c_fmt> <d_fmt> <flags> -%c<dev> -%c<port> -%c<lvl>\n",
  						PrgName, PAR_DEVICE, PAR_PORT, 
  						PAR_DEBUG_LVL			);
  fprintf(stderr,  "\t\t-%c<c_in> -%c<c_out> -%c<d_in> -%c<d_out> -%c<lock>\n",
						PAR_CODE_IN_FILE, PAR_CODE_OUT_FILE, 
						PAR_DATA_IN_FILE, PAR_DATA_OUT_FILE, 
						PAR_LOCKBIT			);
  fprintf(stderr,  "\t\t<c_fmt>        Code fmt: -%c bin, -%c S-Rec, -%c I-Hex\n",
  						PAR_CODE_BIN_FMT, PAR_CODE_SREC_FMT, 
  						PAR_CODE_IHEX_FMT		);
  fprintf(stderr,  "\t\t<d_fmt>        Data fmt: -%c bin, -%c:S-Rec, -%c I-Hex\n",
  						PAR_DATA_BIN_FMT, PAR_DATA_SREC_FMT, 
  						PAR_DATA_IHEX_FMT		);
  fprintf(stderr,  "\t\t<flags>        -%c:erase, -%c:program, -%c:verify\n",
  						PAR_ERASE, PAR_PROG, PAR_VERIFY );
  fprintf(stderr,  "\t\t<dev>          target device. (default: %s), supported:\n",
  						Device->Id);
  Offset = 0;
  do {
    fprintf(stderr,"\t\t\t\t\t");
      Offset = ShowDeviceIds(stderr, 4, Offset);
  } while(Offset != 0);
  fprintf(stderr,  "\t\t<port>         printer port (default: %s), supported:\n",
  						Port->Id);
  Offset = 0;
  do {
    fprintf(stderr,"\t\t\t\t\t");
    Offset = ShowPortIds(stderr, 6, Offset);
  } while(Offset != 0);  
  fprintf(stderr,  "\t\t<lock>         LockBits (supported: ");
  ShowLockBits(stderr);  
  fprintf(stderr,")\n");
  fprintf(stderr,  "\t\t<lvl>          Debug level\n");
  fprintf(stderr,"\tIf no data format given, fmt is determined by file extension\n"); 
}

/*----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------*/
static DEVICE_T *GetDevice(char *Id) {
  DEVICE_T *Dev;
  if(Id == NULL) return(NULL);
  for(Dev = Devices; Dev->Id != NULL; Dev++) {
    if(strcasecmp(Id, Dev->Id) == 0) return(Dev);
  }
  return(NULL);
}

/*----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------*/
static PORT_T *GetPort(char *Id) {
  PORT_T *Pt;
  if(Id == NULL) return(NULL);
  for(Pt = Ports; Pt->Id != NULL; Pt++) {
    if(strcasecmp(Id, Pt->Id) == 0) return(Pt);
  }
  return(NULL);
}

/*----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------*/
static  LOCKBIT_T *GetLockBit(char *Id, char **EndPtr) {
  int i;
  char *End;
  
  if(Id == NULL) return(NULL);
  for(i = 0; LockBits[i] != NULL; i++) {
    if(strncasecmp(Id, LockBits[i]->Id, strlen(LockBits[i]->Id)) == 0) {
      End = Id+strlen(LockBits[i]->Id);
      if((*End == PAR_DELIM) || (*End == '\0')) {
        if(EndPtr != NULL) *EndPtr = *End == '\0' ? NULL : End;
        return(LockBits[i]);
      }
    }
  }
  return(NULL);
}

/*----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------*/
static void AppendLockBit(LOCKBIT_T *LbList[], char *Id) {
  LOCKBIT_T *Lb;  
  int i;

  if((LbList == NULL) || (Id == NULL)) return;
  do{
    if((Lb = GetLockBit(Id, &Id)) == NULL) { 
      fprintf(stderr,"*** Error: Invalid Lockbit '%s'\n", Id);
      exit(3);
    }
    for(i = 0; LbList[i] != NULL; i++) if(LbList[i] == Lb) break; 
    if(LbList[i] == NULL) {
      LbList[i++] = Lb;
      LbList[i]   = NULL;
    }
    if(Id != NULL) {
      Id++;
      if(*Id == '\0') Id = NULL;
    } /* if(Id != NULL) */ 
  }while(Id != NULL);
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------*/
static DATA_T SuggestFmt(char *FileName) {
  char *Extension;
  int i;
  if((Extension = strrchr(FileName, EXT_DELIM)) == NULL) return(UNDEF);
  Extension++;
  for(i = 0; BinExtension[i] != NULL; i++) {
    if(strcasecmp(Extension, BinExtension[i]) == 0) return(BINARY);
  }
  for(i = 0; SRecExtension[i] != NULL; i++) {
    if(strcasecmp(Extension, SRecExtension[i]) == 0) return(MOTO_SREC);
  }
  for(i = 0; IHexExtension[i] != NULL; i++) {
    if(strcasecmp(Extension, IHexExtension[i]) == 0) return(INTEL_HEX);
  }
  return(UNDEF);
} 

/*--------------------------------------------------------------------- 
 *
 *-----------------------------------------------------------------------------*/
static int GetCmdLine(int argc, char *argv[]) {
  int Cmd;
 
  if(argc <= 1) return(-1);
  do{
    if((Cmd = getopt(argc, argv, CmdLine)) != EOF) {
      if((optarg != NULL) && (*optarg != '\0') && (*optarg == '=')) optarg++; 
      switch(Cmd) {
	case PAR_CODE_IN_FILE: 
	    if((CodeIn = fopen(optarg, "r")) == NULL) {
	      fprintf(stderr,"*** Cannot open code input file '%s' ***\n", optarg);
	      return(-1);
	    }
	    SuggestCodeFmt = SuggestFmt(optarg);	
	  break;
	case PAR_CODE_OUT_FILE: 
	    if((CodeOut = open(optarg, O_WRONLY | O_TRUNC | O_CREAT, 	\
	    				S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0){
	      fprintf(stderr,"*** Cannot open code output file '%s' ***\n", optarg);
	      return(-2);
	    }
	  break;
	case PAR_DATA_IN_FILE:
	    if((DataIn = fopen(optarg, "r")) == NULL) {
	      fprintf(stderr,"*** Cannot open data input file '%s' ***\n", optarg);
	      return(-3);
	    }
	    SuggestDataFmt = SuggestFmt(optarg);	
	  break;
	case PAR_DATA_OUT_FILE: 
	    if((DataOut = open(optarg, O_WRONLY | O_TRUNC | O_CREAT, 	\
	    				S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0){
	      fprintf(stderr,"*** Cannot open data output file '%s' ***\n", optarg);
	      return(-4);
	    }
	  break;
	case PAR_ERASE: 	Erase++;  				break;
	case PAR_LOCKBIT: 	AppendLockBit(Lock, optarg);		break;
	case PAR_DEBUG_LVL: 	DebugLevel = strtoul(optarg, NULL, 0); 	break;
	case PAR_PORT:	 	Port = GetPort(optarg);			break;
	case PAR_DEVICE: 	Device = GetDevice(optarg);		break;
	case PAR_VERIFY:	Verify++;				break;
	case PAR_PROG : 	Programm++;				break;
	case PAR_CODE_BIN_FMT:	CodeFmt = BINARY; 			break;
	case PAR_DATA_BIN_FMT:	DataFmt = BINARY;			break;
	case PAR_DATA_IHEX_FMT: DataFmt = INTEL_HEX;			break;
	case PAR_CODE_IHEX_FMT: CodeFmt = INTEL_HEX;			break;
	case PAR_DATA_SREC_FMT: DataFmt = MOTO_SREC;			break;
	case PAR_CODE_SREC_FMT: CodeFmt = MOTO_SREC;			break;
        default: return(-5);
      } /* switch(Cmd) */
    } /* if((Cmd = getopt(argc, argv, CmdLine)) != EOF) */
  } while(Cmd != EOF);
  if(optind != argc) return(-6);
  return(0);
}  

/*------------------------------------------------------------------
 *
 *--------------------------------------------------------------------*/
static int LockBitSupported(LOCKBIT_T *Lk, LOCKBIT_T *List[]) {
  int i;
  for(i = 0; List[i] != NULL; i++) if(Lk == List[i]) return(1);
  return(0); 
}

/*----------------------------------------------------------------
 *
 *------------------------------------------------------------------------*/
static char *DataFmtStr(DATA_T Type) {
  
  switch(Type) {
    case BINARY:      return("binary");
    case MOTO_SREC:   return("moto-s-record");
    case INTEL_HEX:   return("intel-hex");
    default:          return("unknown");
  }
  return("unknown");
}

/*-------------------------------------------------------------------
 *
 *---------------------------------------------------------------------*/
static void ShiftLockBits(LOCKBIT_T *Lk[]) {
  int i;
  for(i = 0; Lk[i] != NULL; i++) if((Lk[i] = Lk[i+1]) == NULL) return;
}
         
/*------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static int CheckPar(void) {
  if(Port == NULL) {
    fprintf(stderr,"*** Error: invalid port selected\n");
    return(-1);
  }
  if(Device == NULL) {
    fprintf(stderr,"*** Error: invalid device selected\n");
    return(-2);
  }
  if((DataIn != NULL) && (DataFmt == UNDEF)) {
    DataFmt = SuggestDataFmt;
    fprintf(stderr,"warning: no data format specified -- suggesting %s\n",
    		DataFmtStr(DataFmt));
  }
  if((CodeIn != NULL) && (CodeFmt == UNDEF)) {
    CodeFmt = SuggestCodeFmt;
    fprintf(stderr,"warning: no code format specified -- suggesting %s\n",
    		DataFmtStr(CodeFmt));
  }
  if((Verify != 0) && (DataIn == NULL) && (CodeIn == NULL)) {
    fprintf(stderr,"*** Error: no input data specified for verifying\n");
    return(-3);
  }
  if((Programm != 0) && (DataIn == NULL) && (CodeIn == NULL)) {
    fprintf(stderr,"*** Error: no input data specified for programming\n");
    return(-4);
  }
  if(((DataIn != NULL) || (CodeIn != NULL)) && (Programm == 0) && (Verify == 0))
  	Programm = 1;
  if(Lock[0] != NULL) {
    int i;
    for(i = 0; Lock[i] != NULL; i++) {
      while(1) {
        if(LockBitSupported(Lock[i], Device->LockBits)) break;
        ShiftLockBits(&(Lock[i]));
        if(Lock[i] == NULL) break;
      } /* while(1) */
    } /* for(i = 0; Lock[i] != NULL; i++) */
  }
  return(0);
}

/*------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static void ShowParameter(void) {
  int i;
 
  printf("Port:          %s\n", Port->Id);
  printf("Device:        %s\n", Device->Id);
  printf("Datenformat:   %s\n", DataFmtStr(DataFmt));
  printf("Codeformat:    %s\n", DataFmtStr(CodeFmt));
  printf("Verify:        %s\n", Verify == 0 ? "Off" : "On");
  printf("Programm:      %s\n", Programm  == 0 ? "Off" : "On");
  printf("Erase:         %s\n", Erase  == 0 ? "Off" : "On");
  printf("Lockbits:      ");
  for(i=0; Lock[i]!=NULL;i++) { 
    printf("%s%c", Lock[i]->Id, Lock[i+1]!=NULL ? ',' : ' ');
  } /* for(i=0; Lock[i]!=NULL;i++) */
  printf("\n");
}

/*--------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static unsigned char *At89Malloc(long Len) {
  unsigned char *Buf;
  
  if((Buf = malloc(sizeof(*Buf) * Len)) == NULL) {
    fprintf(stderr,"*** Malloc Error, allocatin %ld bytes\n", Len);
    exit(100);
  } 
  return(Buf);
} 

/*---------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static void At89Exit(int Code) {
  ReleaseStk200();
  exit(Code);
}

/*-------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int main(int argc, char *argv[]) {
  unsigned char *CodePrgBuf, *CodeReadBuf, *DataPrgBuf, *DataReadBuf;
  long CodeLen = 0, DataLen = 0;
  long ErrCode;
     
  /*------------------------ Programname ------------------------------------*/
  if((PrgName = rindex(argv[0], '/')) != NULL) PrgName++;
  else PrgName = argv[0];
  /*----------- Get command line parameters ---------------------------------*/
  if(GetCmdLine(argc, argv) != 0) {
    Usage();
    exit(1);
  }     
  if(CheckPar() != 0) {
    Usage();
    exit(2);
  }
  ShowParameter();
  /*----------------- demanding port access --------------------------*/  
  printf("Initializing STK200-Adaptor...\n"); 
  fflush(stdout);
  if((ErrCode = InitStk200(Port->Port)) != 0) {
     fprintf(stderr, "\t-> ErrorCode %ld\n", ErrCode);
     exit(1);
  }
  /*----------------- activating stk200 ------------------------------*/
  printf("Activating device...\n");  
  fflush(stdout);
  if((ErrCode = ActivateStk200()) != 0) {
     fprintf(stderr, "\t-> ErrorCode %ld\n", ErrCode);
     exit(2);
  }
  /*---------------- allocate buffers --------------------------------*/
  CodePrgBuf	= At89Malloc(Device->CodeSize);
  CodeReadBuf	= At89Malloc(Device->CodeSize);
  DataPrgBuf	= At89Malloc(Device->DataSize);
  DataReadBuf	= At89Malloc(Device->DataSize);
  memset(CodePrgBuf, BUFFER_INIT_VAL, Device->CodeSize);
  memset(CodeReadBuf,BUFFER_INIT_VAL, Device->CodeSize);
  memset(DataPrgBuf, BUFFER_INIT_VAL, Device->DataSize);
  memset(DataReadBuf,BUFFER_INIT_VAL, Device->DataSize);
  /*---------------- erase -------------------------------------------*/
  if(Erase != 0) {
    ErrCode = ChipErase();
    if(ErrCode != 0) {
      fprintf(stderr,"Error %ld erasing chip\n", ErrCode);
      At89Exit(1);
    }
  }
  /*---------------- loading code and/or data ------------------------*/
  if(DataIn != NULL) {
    DataLen = LoadData(DataIn, DataFmt, 0, Device->DataSize, DataPrgBuf, 1);
    fclose(DataIn);
    if(DataLen <= 0) {
      fprintf(stderr,"*** Error %ld loading Data file\n", DataLen);
      At89Exit(2);
    }
    printf("%ld Data bytes loaded\n", DataLen);
  }
  if(CodeIn != NULL) {
    CodeLen = LoadData(CodeIn, CodeFmt, 0, Device->CodeSize, CodePrgBuf, 1); 
    fclose(CodeIn);
    if(CodeLen <= 0) {
      fprintf(stderr,"*** Error %ld loading Code file\n", CodeLen);
      At89Exit(3);
    }
    printf("%ld Code  bytes loaded\n", CodeLen);
  }
  /*---------------- programming -------------------------------------*/
  if(Programm != 0) {
    if(DataIn != NULL) {
      WriteDataMem(DataPrgBuf,(Erase !=0) ? DataLen:Device->DataSize,Erase,Device);
    }
    if(CodeIn != NULL) {
      WriteCodeMem(CodePrgBuf,(Erase !=0) ? CodeLen:Device->CodeSize,Erase,Device);
    }
  }
  /*---------------- reading back ------------------------------------*/
  if((Verify != 0) | (DataOut >= 0) | (CodeOut >= 0)) {
    ReadCodeMem(CodeReadBuf, Device->CodeSize, Device);
    ReadDataMem(DataReadBuf, Device->DataSize, Device);
  } 
  if((Verify != 0) && (CodeIn != NULL)) {
    ErrCode = memcmp(CodeReadBuf, CodePrgBuf, (size_t)CodeLen);
    printf("Verify Code Rom %s\n", ErrCode == 0 ? "succeded" : "*** failed ***"); 
  } 
  if((Verify != 0) && (DataIn != NULL)) {
    ErrCode = memcmp(DataReadBuf, DataPrgBuf, (size_t)DataLen);
    printf("Verify Data Rom %s\n", ErrCode == 0 ? "succeded" : "*** failed ***"); 
  } 
  if(DataOut >= 0) {
    ErrCode = write(DataOut, DataReadBuf, Device->DataSize);
    close(DataOut);
    if(ErrCode != Device->DataSize) {
      fprintf(stderr,"*** Error: write %ld B of data instead of %ld B\n",
      		ErrCode, Device->DataSize); 
    }
  }
  if(CodeOut >= 0) {
    ErrCode = write(CodeOut, CodeReadBuf, Device->CodeSize);
    close(CodeOut);
    if(ErrCode != Device->CodeSize) {
      fprintf(stderr,"*** Error: write %ld B of code of %ld B\n",
      		ErrCode, Device->CodeSize);
    }
  }
  if(Lock[0] != NULL) {
    WriteLockBits(Lock, Device);
  }
  /*----------------- the end ----------------------------------------*/
  printf("Release STK200-Adaptor...\n");
  fflush(stdout);
  if((ErrCode = ReleaseStk200()) != 0) { 
    DBG_PRINTF(DBG_IO, "\t-> ErrorCode %ld\n", ErrCode);
    exit(2);
  }
  return(0);
}
