/**************************************************************************
 * avr-stk200 -- downloader for ATMEL AVR microcontrollers, using the
 *	STK200 Interface.
 * 
 * 
 *
 *
 */
#define VERSION	"0.1b"
/*
 *
 **************************************************************************/

#include <stdio.h>
#include <string.h>
#include <unistd.h>             /* getopt(), write(), close() */
#include <sys/types.h>          /* open() */
#include <sys/stat.h>           /* open() */
#include <fcntl.h>              /* open() */
#include <stdlib.h>             /* strtoul(), malloc() */
#include <string.h>
#include <ctype.h>              /* atoi() */                     

/*-------------------- local headers ---------------------------------------*/
#ifdef extern
#  undef extern
#endif
#define extern
#  include "avr.h"
#  include "debug.h"
#undef extern

#include "port-io.h"
#include "loaddata.h"

/*-------------------- types -----------------------------------------------*/
typedef unsigned char BYTE_T;

typedef struct sPorts { char *Id;
		unsigned long Port;
	} PORT_T;

/*-------------------- variables -------------------------------------------*/
static PORT_T Ports[] = {
		{ "lp0",	0x378,	},
		{ "lp1",	0x278,	},
		{ "lp2",	0x3bc,	},
		/*--------- End ------------*/
		{ NULL,		0,      }, };

struct sLock {
	LOCKBIT_T	Lb;
	char		*Id;
	char		*Expl;
	} LockId[] =	{
		{LOCK1,		"Lb1",	"Lockbit 1",	},
		{LOCK2,		"Lb2",	"Lockbit 2",	},
		{BOOTLOCK12,	"Bl12",	"Bootlock 12",	},
		{BOOTLOCK11,	"Bl11",	"Bootlock 11",	},
		{BOOTLOCK02,	"Bl02",	"Bootlock 02",	},
		{BOOTLOCK01,	"Bl01",	"Bootlock 01",	},
		/*----------------- end -----------------------*/
		{NOLOCK,		NULL,	NULL,		}, };

struct sFuse {
	FUSEBIT_T	Fb;
	char		*Id;
	char		*Expl;
	} FuseId[] =	{
		{SPIEN,		"SPIEN",  	"SPIEN", 	},        	
		{RCEN,		"RCEN",		"RCEN",		}, 
		{FSTRT,		"FSTRT",	"FSTRT",	}, 
		{BODLEVEL,	"BODLEVEL",	"BODLEVEL"	}, 
		{BODEN,		"BODEN",	"BODEN",	}, 
		{CKSEL3,	"CKSEL3",	"CKSEL3",	}, 
		{CKSEL2,	"CKSEL2",	"CKSEL2",	}, 
		{CKSEL1,	"CKSEL1",	"CKSEL1",	}, 
		{CKSEL0,	"CKSEL0",	"CKSEL0",	}, 
		{EESAVE,	"EESAVE",	"EESAVE",	}, 
		{SUT1,		"SUT1",		"SUT1",		}, 
		{SUT0,		"SUT0",		"SUT0",		}, 
		{BOOTRST,	"BOOTRST",	"BOOTRST",	}, 
		{RSTDSBL,	"RSTDSBL",	"RSTDSBL",	}, 
		/*------------------ end ----------------------------*/
		{NOFUSE,	NULL,		NULL,		}, };
		
typedef int (*RD_WR_FCT)(unsigned char *Buffer, long Len);
typedef int (*LOCK_FCT)(LOCKBIT_T *Lock);
typedef int (*FUSE_FCT)(FUSEBIT_T *Fuse);
typedef int (*CLR_FCT)(void);
		
typedef struct sDevice	{
		char 		*Id;
		long		CodeSize;
		long		DataSize;
		BYTE_T		*Sig;
		int		SigLen;
		CLR_FCT		ChipErase;
		RD_WR_FCT	RdCodeMem;
		RD_WR_FCT	WrCodeMem;
		RD_WR_FCT	RdDataMem;
		RD_WR_FCT	WrDataMem;
		LOCK_FCT	RdLock;
		LOCK_FCT	WrLock;
		FUSE_FCT	RdFuse;
		FUSE_FCT	WrFuse;
		LOCKBIT_T	*LockBits;
		FUSEBIT_T	*FuseBits;
	} DEVICE_T;
		

/*-------------------------------- lock and fuse bits -----------------------*/
static LOCKBIT_T AvrLockBits[] 	= { LOCK1, LOCK2, NOLOCK };
static FUSEBIT_T NoFuse[]	= { NOLOCK };
static FUSEBIT_T Fuse2313[]	= { NOLOCK };
static FUSEBIT_T Fuse2323[]	= { FSTRT, NOLOCK };
static FUSEBIT_T Fuse2343[]	= { RCEN,  NOLOCK };
static FUSEBIT_T Fuse4434_8535[]= { FSTRT, NOLOCK };
static FUSEBIT_T Fuse2333_4433[]= { CKSEL0, CKSEL1, CKSEL2, BODEN, BODLEVEL, NOLOCK };
/*------------------------------- Signatures --------------------------------*/
static BYTE_T Sig1200[] = {0x1E, 0x90, 0x01};
static BYTE_T Sig2313[] = {0x1E, 0x91, 0x01};
static BYTE_T Sig2323[] = {0x1E, 0x91, 0x02};
static BYTE_T Sig2333[] = {0x1E, 0x91, 0x05};
static BYTE_T Sig2343[] = {0x1E, 0x91, 0x03};
static BYTE_T Sig4414[] = {0x1E, 0x92, 0x01};
static BYTE_T Sig4433[] = {0x1E, 0x92, 0x03};
static BYTE_T Sig4434[] = {0x1E, 0x92, 0x02};
static BYTE_T Sig8515[] = {0x1E, 0x93, 0x01};
static BYTE_T Sig8535[] = {0x1E, 0x93, 0x03};

#define SIG_LEN(a)	(sizeof(a)/sizeof(a[0]))
#define MAX_SIG_LEN	(SIG_LEN(Sig1200)+1)

/*-------------------------------- supported devices ---------------------------*/
static DEVICE_T Devices[] = {
	{Id:"AT90S1200",	CodeSize: 1024,	DataSize: 64,	
		Sig: Sig1200,	SigLen: SIG_LEN(Sig1200), 	
		ChipErase:	ChipErase,
		RdCodeMem:	ReadCodeMem,	WrCodeMem: 	WriteCodeMem,
		RdDataMem:	ReadDataMem,	WrDataMem: 	WriteDataMem,
		RdLock:		NULL,		WrLock:	   	WriteLockAvr,
		RdFuse:		NULL,		WrFuse:		NULL,
		LockBits:	AvrLockBits,	FuseBits:	NoFuse,		},

	{Id:"AT90S2313",	CodeSize: 2*1024,DataSize: 128,	
		Sig: Sig2313,	SigLen: SIG_LEN(Sig2313),
		ChipErase:	ChipErase,
		RdCodeMem:	ReadCodeMem,	WrCodeMem: 	WriteCodeMem,
		RdDataMem:	ReadDataMem,	WrDataMem: 	WriteDataMem,
		RdLock:		NULL,		WrLock:	   	WriteLockAvr,
		RdFuse:		NULL,		WrFuse:		NULL,
		LockBits:	AvrLockBits,	FuseBits:	Fuse2313,	},			

	{Id:"AT90S2323",	CodeSize: 2*1024,DataSize: 128,	
		Sig: Sig2323,	SigLen: SIG_LEN(Sig2323),
		ChipErase:	ChipErase,
		RdCodeMem:	ReadCodeMem,	WrCodeMem: 	WriteCodeMem,
		RdDataMem:	ReadDataMem,	WrDataMem: 	WriteDataMem,
		RdLock:		ReadLock2323_2343_4434_8535,	
		WrLock:	   	WriteLockAvr,
		RdFuse:		ReadFuse2323_4434_8535,
		WrFuse:		WriteFuse2323_4434_8535,
		LockBits:	AvrLockBits,	FuseBits:	Fuse2323	},

	{Id:"AT90S2333",	CodeSize: 2*1024,DataSize: 128,	
		Sig: Sig2333,	SigLen: SIG_LEN(Sig2333),
		ChipErase:	ChipErase,
		RdCodeMem:	ReadCodeMem,	WrCodeMem: 	WriteCodeMem,
		RdDataMem:	ReadDataMem,	WrDataMem: 	WriteDataMem,
		RdLock:		ReadLockAvr,	WrLock:	   	WriteLockAvr,
		RdFuse:		ReadFuse2333_4433,
		WrFuse:		WriteFuse2333_4433,
		LockBits:	AvrLockBits,	FuseBits:	Fuse2333_4433	},		

	{Id:"AT90S2343",	CodeSize: 2*1024,DataSize: 128,	
		Sig: Sig2343,	SigLen: SIG_LEN(Sig2343),
		ChipErase:	ChipErase,
		RdCodeMem:	ReadCodeMem,	WrCodeMem: 	WriteCodeMem,
		RdDataMem:	ReadDataMem,	WrDataMem: 	WriteDataMem,
		RdLock:		ReadLock2323_2343_4434_8535,	
		WrLock:	   	WriteLockAvr,
		RdFuse:		ReadFuse2343,
		WrFuse:		WriteFuse2343,
		LockBits:	AvrLockBits,	FuseBits:	Fuse2343	},

	{Id:"AT90S4414",	CodeSize: 4*1024,DataSize: 256,	
		Sig: Sig4414,	SigLen: SIG_LEN(Sig4414),
		ChipErase:	ChipErase,
		RdCodeMem:	ReadCodeMem,	WrCodeMem: 	WriteCodeMem,
		RdDataMem:	ReadDataMem,	WrDataMem: 	WriteDataMem,
		RdLock:		NULL,		WrLock:	   	WriteLockAvr,
		RdFuse:		NULL,		WrFuse:		NULL,
		LockBits:	AvrLockBits,	FuseBits:	NoFuse	},

	{Id:"AT90S4433",	CodeSize: 4*1024,DataSize: 256,	
		Sig: Sig4433,	SigLen: SIG_LEN(Sig4433),
		ChipErase:	ChipErase,
		RdCodeMem:	ReadCodeMem,	WrCodeMem: 	WriteCodeMem,
		RdDataMem:	ReadDataMem,	WrDataMem: 	WriteDataMem,
		RdLock:		ReadLockAvr,	WrLock:	   	WriteLockAvr,
		RdFuse:		ReadFuse2333_4433,
		WrFuse:		WriteFuse2333_4433,
		LockBits:	AvrLockBits,	FuseBits:	Fuse2333_4433	},

	{Id:"AT90S4434",	CodeSize: 4*1024,DataSize: 256,	
		Sig: Sig4434,	SigLen: SIG_LEN(Sig4434),
		ChipErase:	ChipErase,
		RdCodeMem:	ReadCodeMem,	WrCodeMem: 	WriteCodeMem,
		RdDataMem:	ReadDataMem,	WrDataMem: 	WriteDataMem,
		RdLock:		ReadLock2323_2343_4434_8535,	
		WrLock:	   	WriteLockAvr,
		RdFuse:		ReadFuse2323_4434_8535,
		WrFuse:		WriteFuse2323_4434_8535,
		LockBits:	AvrLockBits,	FuseBits:	Fuse4434_8535	},

	{Id:"AT90S8515",	CodeSize: 8*1024,DataSize: 512,	
		Sig: Sig8515,	SigLen: SIG_LEN(Sig8515),
		ChipErase:	ChipErase,
		RdCodeMem:	ReadCodeMem,	WrCodeMem: 	WriteCodeMem,
		RdDataMem:	ReadDataMem,	WrDataMem: 	WriteDataMem,
		RdLock:		NULL,		WrLock:	   	WriteLockAvr,
		RdFuse:		NULL,		WrFuse:		NULL,
		LockBits:	AvrLockBits,	FuseBits:	NoFuse		},

	{Id:"AT90S8535",	CodeSize: 8*1024,DataSize: 512,	
		Sig: Sig8535,	SigLen: SIG_LEN(Sig8535),
		ChipErase:	ChipErase,
		RdCodeMem:	ReadCodeMem,	WrCodeMem: 	WriteCodeMem,
		RdDataMem:	ReadDataMem,	WrDataMem: 	WriteDataMem,
		RdLock:		ReadLock2323_2343_4434_8535,	
		WrLock:	   	WriteLockAvr,
		RdFuse:		ReadFuse2323_4434_8535,
		WrFuse:		WriteFuse2323_4434_8535,
		LockBits:	AvrLockBits,	FuseBits:	Fuse4434_8535	},
	/*-------------------- end ---------------------------------------------*/
	{Id:NULL,		0,		0		},		};

/*-------------------- defines -------------------------------------------------*/
#define EXT_DELIM	'.'
#define PAR_DELIM	','

static char *BinExtension[] = { 
		"bin", 
		"rom", 
		"img",
		/*------*/
		NULL    };
static char *SRecExtension[] = {
		"srec",
		"mot",
		"moto",
		"rec",
		/*-----*/
		NULL    };
static char *IHexExtension[] = {
		"ihex",
		"hex",
		"ihx",
		"int",
		"intel",
		/*-----*/
		NULL    };      
	
/*-------------------- Command line parameters --------------------------*/
#define PAR_FORCE_DEV		'A'

#define PAR_CODE_BIN_FMT	'b'
#define PAR_DATA_BIN_FMT	'B'

#define PAR_CODE_IN_FILE	'c'
#define PAR_CODE_OUT_FILE	'C'

#define PAR_DATA_IN_FILE	'd'
#define PAR_DATA_OUT_FILE	'D'

#define PAR_ERASE		'e'

#define PAR_WRITE_FUSE		'f'
#define PAR_READ_FUSE		'F'

#define PAR_CODE_IHEX_FMT	'i'
#define PAR_DATA_IHEX_FMT	'I'

#define PAR_WRITE_LOCK		'l'
#define PAR_READ_LOCK 		'L'

#define PAR_PROG		'p'
#define PAR_PORT		'P'

#define PAR_CODE_SREC_FMT	's'
#define PAR_DATA_SREC_FMT	'S'

#define PAR_DEVICE		't'

#define PAR_VERIFY		'v'
#define PAR_DEBUG_LVL		'V'


static char CmdLine[] = {       PAR_CODE_IN_FILE,	':',
				PAR_CODE_OUT_FILE,	':',
				PAR_DATA_IN_FILE,	':',
				PAR_DATA_OUT_FILE,	':',
				PAR_ERASE,
				PAR_WRITE_LOCK,		':',
				PAR_READ_LOCK,
				PAR_WRITE_FUSE,		':',
				PAR_READ_FUSE,
				PAR_DEBUG_LVL,		':',
				PAR_PORT,		':',
				PAR_DEVICE,		':',
				PAR_VERIFY,
				PAR_PROG,
				PAR_CODE_BIN_FMT,
				PAR_DATA_BIN_FMT,
				PAR_DATA_IHEX_FMT,
				PAR_CODE_IHEX_FMT,
				PAR_DATA_SREC_FMT,
				PAR_CODE_SREC_FMT,
				PAR_FORCE_DEV,
				/*------------- end --------------------*/
				'\0'				    };

/*----------------------------- variables ----------------------------*/
static char		*PrgName;
static PORT_T		*Port		= Ports+0;
static DEVICE_T		*Device		= NULL;
static FILE		*DataIn 	= NULL;
static FILE		*CodeIn 	= NULL;
static int		DataOut		= -1;
static int		CodeOut		= -1;
static DATA_T		DataFmt		= UNDEF;
static DATA_T		CodeFmt		= UNDEF;
static DATA_T		SuggestDataFmt	= UNDEF;
static DATA_T		SuggestCodeFmt	= UNDEF;
static int		Verify		= 0;
static int		Programm	= 0;
static int		Erase		= 0;
static int		ForceDev	= 0;
static LOCKBIT_T	Lock[sizeof(LockId)/sizeof(LockId[0])] = { NOLOCK, };
static int 		ReadLock	= 0;
static FUSEBIT_T	Fuse[sizeof(FuseId)/sizeof(FuseId[0])] = { NOFUSE, };
static int 		PrgFuse		= 0;	
static int		ReadFuse	= 0;

/*----------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
static int ShowDeviceIds(FILE *Fp, int Num, int Offs) {
  int i;
    
  for(i = Offs; Devices[i].Id != NULL; i++) {
    fprintf(Fp,"%s",Devices[i].Id);
    if(Devices[i+1].Id == NULL) {
      fprintf(Fp,"\n");
      return(0);
    } else {
      if(i == Offs+Num-1) { 
        fprintf(Fp,"\n");
        return(Num + Offs);
      } else {
        fprintf(Fp,",");
      }
    } /* if(Devices[i+1].Id == NULL) */
  } /* for(i = Offs; Devices[i].Id != NULL; i++) */
  return(0);
}
  
/*----------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
static int ShowPortIds(FILE *Fp, int Num, int Offs) {
  int i;

  for(i = Offs; Ports[i].Id != NULL; i++) {
    fprintf(Fp,"%s",Ports[i].Id);
    if(Ports[i+1].Id == NULL) {
      fprintf(Fp,"\n");
      return(0);
    } else {
      if(i == Offs+Num-1) { 
        fprintf(Fp,"\n");
        return(Num + Offs);
      } else {
        fprintf(Fp,",");
      } 
    } /* else -- if(Ports[i+1].Id == NULL) */ 
  } /* for(i = Offs; Ports[i].Id != NULL; i++) */
  return(0);
}

/*----------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
static int ShowLockBits(FILE *Fp, int Num, int Offs) {
  int i;
  for(i = Offs; LockId[i].Id != NULL; i++) { 
    fprintf(Fp, "%s", LockId[i].Id);
    if(LockId[i+1].Id == NULL) { 
      fprintf(Fp, "\n");
      return(0);
    } else {
      if(i == Offs+Num-1) {
        fprintf(Fp,"\n");
        return(Num + Offs);
      } else {
        fprintf(Fp,",");
      }
    }
  }
  return(0);
}

/*----------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
static int ShowFuseBits(FILE *Fp, int Num, int Offs) {
  int i;
  for(i = Offs; FuseId[i].Id != NULL; i++) { 
    fprintf(Fp, "%s", FuseId[i].Id);
    if(FuseId[i+1].Id == NULL) { 
      fprintf(Fp, "\n");
      return(0);
    } else {
      if(i == Offs+Num-1) {
        fprintf(Fp,"\n");
        return(Num + Offs);
      } else {
        fprintf(Fp,",");
      }
    }
  }
  return(0);
}

/*----------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
static void Usage(void) {
  int Offset = 0;

  fprintf(stderr,  "Usage: %s <c_fmt> <d_fmt> <flags> -%c<dev> -%c<port> "
# ifdef DEBUG  
  			"-%c<lvl>"
# endif
  			"\n",	PrgName, PAR_DEVICE, PAR_PORT	
# ifdef DEBUG 
  						, PAR_DEBUG_LVL	
# endif  					
			);
  fprintf(stderr,  "\t\t-%c<Cin> -%c<Cout> -%c<Din> -%c<Dout> -%c<lock> -%c<fuse>\n",
						PAR_CODE_IN_FILE, PAR_CODE_OUT_FILE, 
						PAR_DATA_IN_FILE, PAR_DATA_OUT_FILE, 
						PAR_WRITE_LOCK,PAR_WRITE_FUSE		);
  fprintf(stderr,  "\t\t<c_fmt>        Code fmt: -%c bin, -%c S-Rec, -%c I-Hex\n",
  						PAR_CODE_BIN_FMT, PAR_CODE_SREC_FMT, 
  						PAR_CODE_IHEX_FMT		);
  fprintf(stderr,  "\t\t<d_fmt>        Data fmt: -%c bin, -%c:S-Rec, -%c I-Hex\n",
  						PAR_DATA_BIN_FMT, PAR_DATA_SREC_FMT, 
  						PAR_DATA_IHEX_FMT		);
  fprintf(stderr,  "\t\t<flags>        -%c:erase, -%c:program, -%c:verify\n",
  						PAR_ERASE, PAR_PROG, PAR_VERIFY);
  fprintf(stderr,  "\t\t\t\t\t-%c:show lock bits, -%c: show fuse bits\n",
  						PAR_READ_LOCK, PAR_READ_FUSE);
  fprintf(stderr,  "\t\t\t\t\t-%c:force dev\n", PAR_FORCE_DEV);
  fprintf(stderr,  "\t\t<dev>          target device -- supported devices:\n");
  Offset = 0;
  do {
    fprintf(stderr,"\t\t\t\t\t");
      Offset = ShowDeviceIds(stderr, 4, Offset);
  } while(Offset != 0);
  fprintf(stderr,  "\t\t<port>         printer port (default: %s), supported ports:\n",
  						Port->Id);
  Offset = 0;
  do {
    fprintf(stderr,"\t\t\t\t\t");
    Offset = ShowPortIds(stderr, 6, Offset);
  } while(Offset != 0);  
  fprintf(stderr,  "\t\t<lock>         LockBits -- supported LockBits:\n");
  Offset = 0;
  do {
    fprintf(stderr,  "\t\t\t\t\t");
    Offset = ShowLockBits(stderr, 6, Offset);
  } while(Offset != 0);
  fprintf(stderr,  "\t\t<fuse>         FuseBits -- supported FuseBits:\n");
  Offset = 0;
  do {
    fprintf(stderr,  "\t\t\t\t\t");
    Offset = ShowFuseBits(stderr, 6, Offset);
  } while(Offset != 0);
  fprintf(stderr,  "\t\t\t\tUse -%c= for resetting fuse bits\n", PAR_WRITE_FUSE);
# ifdef DEBUG
  fprintf(stderr,  "\t\t<lvl>          Debug level\n");
# endif
  fprintf(stderr,"\tIf no data format given, fmt is determined by file extension\n"); 
}

/*----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------*/
static DEVICE_T *GetDevice(char *Id) {
  DEVICE_T *Dev;
  if(Id == NULL) return(NULL);
  for(Dev = Devices; Dev->Id != NULL; Dev++) {
    if(strcasecmp(Id, Dev->Id) == 0) return(Dev);
  }
  return(NULL);
}

/*----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------*/
static PORT_T *GetPort(char *Id) {
  PORT_T *Pt;
  if(Id == NULL) return(NULL);
  for(Pt = Ports; Pt->Id != NULL; Pt++) {
    if(strcasecmp(Id, Pt->Id) == 0) return(Pt);
  }
  return(NULL);
}

/*----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------*/
static  LOCKBIT_T GetLockBit(char *Id, char **EndPtr) {
  int i;
  char *End;
  
  if(Id == NULL) return(NOLOCK);
  for(i = 0; LockId[i].Id != NULL; i++) {
    if(strncasecmp(Id, LockId[i].Id, strlen(LockId[i].Id)) == 0) {
      End = Id+strlen(LockId[i].Id);
      if((*End == PAR_DELIM) || (*End == '\0')) {
        if(EndPtr != NULL) *EndPtr = *End == '\0' ? NULL : End;
        return(LockId[i].Lb);
      }
    }
  }
  return(NOLOCK);
}

/*----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------*/
static void AppendLockBit(LOCKBIT_T *List, char *Id) {
  LOCKBIT_T Lb;  
  int i;

  if((List == NULL) || (Id == NULL)) return;
  do{
    if((Lb = GetLockBit(Id, &Id)) == NOLOCK) { 
      fprintf(stderr,"*** Error: Invalid Lockbit '%s'\n", Id);
      exit(3);
    }
    for(i = 0; List[i] != NOLOCK; i++) if(List[i] == Lb) break; 
    if(List[i] == NOLOCK) {
      List[i++] = Lb;
      List[i]   = NOLOCK;
    }
    if(Id != NULL) {
      Id++;
      if(*Id == '\0') Id = NULL;
    } /* if(Id != NULL) */ 
  }while(Id != NULL);
}


/*----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------*/
static  FUSEBIT_T GetFuseBit(char *Id, char **EndPtr) {
  int i;
  char *End;
  
  if(Id == NULL) return(NOFUSE);
  for(i = 0; FuseId[i].Id != NULL; i++) {
    if(strncasecmp(Id, FuseId[i].Id, strlen(FuseId[i].Id)) == 0) {
      End = Id+strlen(FuseId[i].Id);
      if((*End == PAR_DELIM) || (*End == '\0')) {
        if(EndPtr != NULL) *EndPtr = *End == '\0' ? NULL : End;
        return(FuseId[i].Fb);
      }
    }
  }
  return(NOFUSE);
}

/*----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------*/
static void AppendFuseBit(FUSEBIT_T *List, char *Id) {
  FUSEBIT_T Lb;  
  int i;

  if((List == NULL) || (Id == NULL)) return;
  do{
    if((Lb = GetFuseBit(Id, &Id)) == NOFUSE) { 
      fprintf(stderr,"*** Error: Invalid Fusebit '%s'\n", Id);
      exit(3);
    }
    for(i = 0; List[i] != NOFUSE; i++) if(List[i] == Lb) break; 
    if(List[i] == NOFUSE) {
      List[i++] = Lb;
      List[i]   = NOFUSE;
    }
    if(Id != NULL) {
      Id++;
      if(*Id == '\0') Id = NULL;
    } /* if(Id != NULL) */ 
  }while(Id != NULL);
}

/*-----------------------------------------------------------------------
 *
 *-----------------------------------------------------------------------------*/
static DATA_T SuggestFmt(char *FileName) {
  char *Extension;
  int i;
  if((Extension = strrchr(FileName, EXT_DELIM)) == NULL) return(UNDEF);
  Extension++;
  for(i = 0; BinExtension[i] != NULL; i++) {
    if(strcasecmp(Extension, BinExtension[i]) == 0) return(BINARY);
  }
  for(i = 0; SRecExtension[i] != NULL; i++) {
    if(strcasecmp(Extension, SRecExtension[i]) == 0) return(MOTO_SREC);
  }
  for(i = 0; IHexExtension[i] != NULL; i++) {
    if(strcasecmp(Extension, IHexExtension[i]) == 0) return(INTEL_HEX);
  }
  return(UNDEF);
} 

/*--------------------------------------------------------------------- 
 *
 *-----------------------------------------------------------------------------*/
static int GetCmdLine(int argc, char *argv[]) {
  int Cmd;
 
  if(argc <= 1) return(-1);
  do{
    if((Cmd = getopt(argc, argv, CmdLine)) != EOF) {
      if((optarg != NULL) && (*optarg != '\0') && (*optarg == '=')) optarg++; 
      switch(Cmd) {
	case PAR_CODE_IN_FILE: 
	    if((CodeIn = fopen(optarg, "r")) == NULL) {
	      fprintf(stderr,"*** Cannot open code input file '%s' ***\n", optarg);
	      return(-1);
	    }
	    SuggestCodeFmt = SuggestFmt(optarg);	
	  break;
	case PAR_CODE_OUT_FILE: 
	    if((CodeOut = open(optarg, O_WRONLY | O_TRUNC | O_CREAT, 	\
	    				S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0){
	      fprintf(stderr,"*** Cannot open code output file '%s' ***\n", optarg);
	      return(-2);
	    }
	  break;
	case PAR_DATA_IN_FILE:
	    if((DataIn = fopen(optarg, "r")) == NULL) {
	      fprintf(stderr,"*** Cannot open data input file '%s' ***\n", optarg);
	      return(-3);
	    }
	    SuggestDataFmt = SuggestFmt(optarg);	
	  break;
	case PAR_DATA_OUT_FILE: 
	    if((DataOut = open(optarg, O_WRONLY | O_TRUNC | O_CREAT, 	\
	    				S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0){
	      fprintf(stderr,"*** Cannot open data output file '%s' ***\n", optarg);
	      return(-4);
	    }
	  break;
	case PAR_WRITE_FUSE: 	
                if((optarg != NULL) && (*optarg != '\0')) {
		  AppendFuseBit(Fuse, optarg);	
		}	
		PrgFuse++;
	  break;
	case PAR_ERASE: 	Erase++;  				break;
	case PAR_WRITE_LOCK: 	AppendLockBit(Lock, optarg);		break;
	case PAR_DEBUG_LVL: 	DebugLevel = strtoul(optarg, NULL, 0); 	break;
	case PAR_PORT:	 	Port = GetPort(optarg);			break;
	case PAR_DEVICE: 	Device = GetDevice(optarg);		break;
	case PAR_VERIFY:	Verify++;				break;
	case PAR_PROG : 	Programm++;				break;
	case PAR_CODE_BIN_FMT:	CodeFmt = BINARY; 			break;
	case PAR_DATA_BIN_FMT:	DataFmt = BINARY;			break;
	case PAR_DATA_IHEX_FMT: DataFmt = INTEL_HEX;			break;
	case PAR_CODE_IHEX_FMT: CodeFmt = INTEL_HEX;			break;
	case PAR_DATA_SREC_FMT: DataFmt = MOTO_SREC;			break;
	case PAR_CODE_SREC_FMT: CodeFmt = MOTO_SREC;			break;
	case PAR_READ_LOCK:	ReadLock++;				break; 
	case PAR_READ_FUSE:	ReadFuse++;				break;
	case PAR_FORCE_DEV:	ForceDev++;				break;
        default: return(-5);
      } /* switch(Cmd) */
    } /* if((Cmd = getopt(argc, argv, CmdLine)) != EOF) */
  } while(Cmd != EOF);
  if(optind != argc) return(-6);
  return(0);
}  

/*------------------------------------------------------------------
 *
 *--------------------------------------------------------------------*/
static int LockBitSupported(LOCKBIT_T Lk, LOCKBIT_T List[]) {
  int i;
  for(i = 0; List[i] != NOLOCK; i++) if(Lk == List[i]) return(1);
  return(0); 
}

/*------------------------------------------------------------------
 *
 *--------------------------------------------------------------------*/
static int FuseBitSupported(FUSEBIT_T Lk, FUSEBIT_T List[]) {
  int i;
  for(i = 0; List[i] != NOFUSE; i++) if(Lk == List[i]) return(1);
  return(0); 
}

/*----------------------------------------------------------------
 *
 *------------------------------------------------------------------------*/
static char *DataFmtStr(DATA_T Type) {
  
  switch(Type) {
    case BINARY:      return("binary");
    case MOTO_SREC:   return("moto-s-record");
    case INTEL_HEX:   return("intel-hex");
    default:          return("unknown");
  }
  return("unknown");
}

/*-------------------------------------------------------------------
 *
 *---------------------------------------------------------------------*/
static void ShiftLockBits(LOCKBIT_T *Lk) {
  int i;
  for(i = 0; Lk[i] != NOLOCK; i++) if((Lk[i] = Lk[i+1]) == NOLOCK) return;
}

/*-------------------------------------------------------------------
 *
 *---------------------------------------------------------------------*/
static void ShiftFuseBits(FUSEBIT_T *Lk) {
  int i;
  for(i = 0; Lk[i] != NOFUSE; i++) if((Lk[i] = Lk[i+1]) == NOFUSE) return;
}

/*----------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
static char *GetLockBitId(LOCKBIT_T Bit) {
  int i;
  for(i = 0; LockId[i].Id != NULL; i++) if(LockId[i].Lb == Bit) return(LockId[i].Id); 
  return("<unknown>");
}

/*----------------------------------------------------------------------
 *
 *----------------------------------------------------------------------------*/
static char *GetFuseBitId(FUSEBIT_T Bit) {
  int i;
  for(i = 0; FuseId[i].Id != NULL; i++) if(FuseId[i].Fb == Bit) return(FuseId[i].Id); 
  return("<unknown>");
}

/*------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static int CheckPar(void) {
  if(Port == NULL) {
    fprintf(stderr,"*** Error: invalid port selected\n");
    return(-1);
  }
  if((DataIn != NULL) && (DataFmt == UNDEF)) {
    DataFmt = SuggestDataFmt;
    fprintf(stderr,"warning: no data format specified -- suggesting %s\n",
    		DataFmtStr(DataFmt));
  }
  if((CodeIn != NULL) && (CodeFmt == UNDEF)) {
    CodeFmt = SuggestCodeFmt;
    fprintf(stderr,"warning: no code format specified -- suggesting %s\n",
    		DataFmtStr(CodeFmt));
  }
  if((Verify != 0) && (DataIn == NULL) && (CodeIn == NULL)) {
    fprintf(stderr,"*** Error: no input data specified for verifying\n");
    return(-3);
  }
  if((Programm != 0) && (DataIn == NULL) && (CodeIn == NULL)) {
    fprintf(stderr,"*** Error: no input data specified for programming\n");
    return(-4);
  }
  if(((DataIn != NULL) || (CodeIn != NULL)) && (Programm == 0) && (Verify == 0))
  	Programm = 1;

  return(0);
}


/*------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static void ShowParameter(void) {
  int i;

  if(Port != NULL)  
  printf("Port:          %s\n", Port->Id);
  if(Device != NULL)
  printf("Device:        %s\n", Device->Id);
  printf("Datenformat:   %s\n", DataFmtStr(DataFmt));
  printf("Codeformat:    %s\n", DataFmtStr(CodeFmt));
  printf("Verify:        %s\n", Verify == 0 ? "Off" : "On");
  printf("Programm:      %s\n", Programm  == 0 ? "Off" : "On");
  printf("Erase:         %s\n", Erase  == 0 ? "Off" : "On");
  printf("Lockbits:      ");
  if(Lock[0]==NOLOCK) {
    printf("<none>\n");
  } else {
    for(i=0; Lock[i]!=NOLOCK;i++) { 
      printf("%s%c", GetLockBitId(Lock[i]), Lock[i+1]!=NOLOCK ? ',' : '\n');
    } /* for(i=0; Lock[i]!=NOLOCK;i++) */
  }
  printf("Fusebits:      ");
  if(Fuse[0]==NOFUSE) {
    if(PrgFuse == 0) {
      printf("<none>\n");
    } else {
      printf("<clear>\n");
    }
  } else {
    for(i=0; Fuse[i]!=NOFUSE;i++) { 
      printf("%s%c", GetFuseBitId(Fuse[i]), Fuse[i+1]!=NOFUSE ? ',' : ' ');
    } /* for(i=0; Fuse[i]!=NOFUSE;i++) */  
  }
}

/*--------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static unsigned char *AvrMalloc(long Len) {
  unsigned char *Buf;
  
  if((Buf = malloc(sizeof(*Buf) * Len)) == NULL) {
    fprintf(stderr,"*** Malloc Error, allocatin %ld bytes\n", Len);
    exit(100);
  } 
  return(Buf);
} 

/*---------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static void AvrExit(int Code) {
  ReleaseStk200();
  exit(Code);
}

/*---------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static DEVICE_T *GetSigDev(BYTE_T *Sig) {
  int i;

  for(i = 0; Devices[i].Id != NULL; i++) {
    if(memcmp(Devices[i].Sig, Sig, (size_t)Devices[i].SigLen) == 0) 
    	return(&(Devices[i]));
  }
  return(NULL);
}

/*---------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
static void PrintSig(BYTE_T *Sig, int Len) {
  int i;

  for(i = 0; i < Len; i++) 
  	printf("0x%02X%c", Sig[i], (i < Len-1) ? ',': '\n');
}
  
  
/*-------------------------------------------------------------------------
 *
 *---------------------------------------------------------------------------*/
int main(int argc, char *argv[]) {
  unsigned char *CodePrgBuf, *CodeReadBuf, *DataPrgBuf, *DataReadBuf;
  long CodeSize = 0, DataSize = 0;
  long ErrCode;
  int i;
  BYTE_T Sig[MAX_SIG_LEN];
  int SigLen; 
     
  /*------------------------ Programname ------------------------------------*/
  if((PrgName = rindex(argv[0], '/')) != NULL) PrgName++;
  else PrgName = argv[0];
  printf("%s: Ver:" VERSION "(Built:" __DATE__ "/" __TIME__ ") V.Bosch 8/2000\n",
  		PrgName);
  /*----------- Get command line parameters ---------------------------------*/
  if(GetCmdLine(argc, argv) != 0) {
    Usage();
    exit(1);
  }     
  if(CheckPar() != 0) {
    Usage();
    exit(2);
  }

  /*----------------- demanding port access --------------------------*/  
  printf("Initializing STK200-Adaptor...\n"); 
  fflush(stdout);
  if((ErrCode = InitStk200(Port->Port)) != 0) {
     fprintf(stderr, "\t-> ErrorCode %ld\n", ErrCode);
     exit(1);
  }
  /*----------------- activating stk200 ------------------------------*/
  printf("Activating device...\n");  
  fflush(stdout);
  if((ErrCode = ActivateStk200()) != 0) {
     fprintf(stderr, "\t-> ErrorCode %ld\n", ErrCode);
     exit(2);
  }
  /*---------------- reading device signature ------------------------*/
  printf("reading device signature...\n"); fflush(stdout);
  SigLen = ReadSignature(Sig, MAX_SIG_LEN);
  printf("Signature target device: ");
  PrintSig(Sig, SigLen);
  if(Device == NULL) {
    if((Device = GetSigDev(Sig)) == NULL) {
      fprintf(stderr,"*** error: unknown device -- "
      		"choose similar and use -%c flag\n", PAR_FORCE_DEV); 
      AvrExit(1); 
    } /* if((Device = GetSigDev(Sig)) == NULL) */
  } else { /* if(Device == NULL) */
    if(ForceDev == 0) {
      if(memcmp(Device->Sig, Sig, (size_t)Device->SigLen) != 0) {
        DEVICE_T *NewDev;
        fprintf(stderr,"*** Error: read wrong signature:\n");
        fprintf(stderr,"\tchoosen device (%s) signature: ", Device->Id); 
        PrintSig(Device->Sig, Device->SigLen);
        if((NewDev = GetSigDev(Sig)) != NULL) {
          fprintf(stderr,"\tsuggested device (%s) signature: ", NewDev->Id);
          PrintSig(NewDev->Sig, NewDev->SigLen);
        } /* if((NewDev = GetSigDev(Sig)) != NULL) */
        AvrExit(1);
      } /* if(memcmp(Device->Sig, Sig, (size_t)Device->SigLen) != 0 */
    } /* if(ForceDev == 0) */
  } /* else -- if(Device == NULL) */
  /*------------------------- check lock and fuse bits --------------------------*/  
  if(Lock[0] != NOLOCK) {
    for(i = 0; Lock[i] != NOLOCK; i++) {
      while(1) {
        if(LockBitSupported(Lock[i], Device->LockBits)) break;
        fprintf(stderr,"warning: Lockbit '%s' is not supported by the device\n",
        		GetLockBitId(Lock[i]));
        ShiftLockBits(&(Lock[i]));
        if(Lock[i] == NOLOCK) break;
      } /* while(1) */
    } /* for(i = 0; Lock[i] != NOLOCK; i++) */
  }
  if(Fuse[0] != NOFUSE) {
    for(i = 0; Fuse[i] != NOFUSE; i++) {
      while(1) {
        if(FuseBitSupported(Fuse[i], Device->FuseBits)) break;
        fprintf(stderr,"warning: Fusebit '%s' is not supported by the device\n",
        		GetFuseBitId(Fuse[i]));
        ShiftFuseBits(&(Fuse[i]));
        if(Fuse[i] == NOFUSE) break;
      } /* while(1) */
    } /* for(i = 0; Fuse[i] != NOFUSE; i++) */
  }  
  ShowParameter();
  /*---------------- allocate buffers --------------------------------*/
  CodePrgBuf	= AvrMalloc(Device->CodeSize);
  CodeReadBuf	= AvrMalloc(Device->CodeSize);
  DataPrgBuf	= AvrMalloc(Device->DataSize);
  DataReadBuf	= AvrMalloc(Device->DataSize);
  memset(CodePrgBuf, BUFFER_INIT_VAL, Device->CodeSize);
  memset(CodeReadBuf,BUFFER_INIT_VAL, Device->CodeSize);
  memset(DataPrgBuf, BUFFER_INIT_VAL, Device->DataSize);
  memset(DataReadBuf,BUFFER_INIT_VAL, Device->DataSize);
  /*---------------- erase -------------------------------------------*/
  if(Erase != 0) {
    printf("erasing device..."); fflush(stdout);
    ErrCode = ChipErase();
    printf("done\n"); fflush(stdout);
    if(ErrCode != 0) {
      fprintf(stderr,"Error %ld erasing chip\n", ErrCode);
      AvrExit(1);
    }
  }
  /*---------------- loading code and/or data ------------------------*/
  if(DataIn != NULL) {
    printf("loading data..."); fflush(stdout);
    DataSize = LoadData(DataIn, DataFmt, 0, Device->DataSize, DataPrgBuf, 1);
    fclose(DataIn);
    if(DataSize <= 0) {
      fprintf(stderr,"*** Error %ld loading Data file\n", DataSize);
      AvrExit(2);
    }
    printf("%ld Data bytes loaded\n", DataSize); fflush(stdout);
  }
  if(CodeIn != NULL) {
    printf("loading code..."); fflush(stdout);
    CodeSize = LoadData(CodeIn, CodeFmt, 0, Device->CodeSize, CodePrgBuf, 1); 
    fclose(CodeIn);
    if(CodeSize <= 0) {
      fprintf(stderr,"*** Error %ld loading Code file\n", CodeSize);
      AvrExit(3);
    }
    printf("%ld Code  bytes loaded\n", CodeSize); fflush(stdout);
  }
  /*---------------- programming -------------------------------------*/
  if(Programm != 0) {
    if(DataIn != NULL) {
      if(Device->WrDataMem == NULL) {
        fprintf(stderr,"*** error: no data writing function for '%s' defined\n",
        		Device->Id);
      } else {
        printf("programming %ld bytes of data memory...", DataSize); fflush(stdout);
        Device->WrDataMem(DataPrgBuf, DataSize);
        printf("done\n"); fflush(stdout);
      }
    }
    if(CodeIn != NULL) {
      if(Device->WrCodeMem == NULL) {
        fprintf(stderr,"*** error: no code writing function for '%s' defined\n",
        		Device->Id);
      } else {
        printf("programming %ld bytes of code memory...", CodeSize); fflush(stdout);
        Device->WrCodeMem(CodePrgBuf, CodeSize);
        printf("done\n"); fflush(stdout);
      }
    }
  }
  /*---------------- reading back ------------------------------------*/
  if((Verify != 0) | (DataOut >= 0) | (CodeOut >= 0)) {
    if(Device->RdCodeMem != NULL) {
      printf("reading code memory..."); fflush(stdout);
      ReadCodeMem(CodeReadBuf, Device->CodeSize);
      printf("done\n"); fflush(stdout);
    }
    if(Device->RdDataMem != NULL) {
      printf("reading data memory..."); fflush(stdout);
      ReadDataMem(DataReadBuf, Device->DataSize);
      printf("done\n"); fflush(stdout);
    }
  } 
  if((Verify != 0) && (CodeIn != NULL) && (Device->RdCodeMem != NULL)) {
    ErrCode = memcmp(CodeReadBuf, CodePrgBuf, (size_t)Device->CodeSize);
    printf("Verify Code Rom %s\n", ErrCode == 0 ? "succeded" : "*** failed ***"); 
  } 
  if((Verify != 0) && (DataIn != NULL) && (Device->RdDataMem != NULL)) {
    ErrCode = memcmp(DataReadBuf, DataPrgBuf, (size_t)Device->DataSize);
    printf("Verify Data Rom %s\n", ErrCode == 0 ? "succeded" : "*** failed ***"); 
  } 
  if((DataOut >= 0) && (Device->RdDataMem != NULL)) {
    ErrCode = write(DataOut, DataReadBuf, Device->DataSize);
    close(DataOut);
    if(ErrCode != Device->DataSize) {
      fprintf(stderr,"*** Error: write %ld B of data instead of %ld B\n",
      		ErrCode, Device->DataSize); 
    }
  }
  if((CodeOut >= 0) && (Device->RdCodeMem != NULL)) {
    ErrCode = write(CodeOut, CodeReadBuf, Device->CodeSize);
    close(CodeOut);
    if(ErrCode != Device->CodeSize) {
      fprintf(stderr,"*** Error: write %ld B of code of %ld B\n",
      		ErrCode, Device->CodeSize);
    }
  }
  /*----------------- write lock bits -------------------------------*/
  if(Lock[0] != NOLOCK) {
    if(Device->WrLock == NULL) {
      fprintf(stderr,"*** no function for writing lock bits defined\n");
    } else {
      printf("writing lock bits.."); fflush(stdout);
      Device->WrLock(Lock);
      printf("done\n"); fflush(stdout);
    }
  }
  /*----------------- write fuse bits -------------------------------*/
  if(PrgFuse != 0) {
    if(Device->WrFuse == NULL) {
      fprintf(stderr,"*** no function for writing fuse bits defined\n");
    } else {
      printf("writing fuse bits.."); fflush(stdout);
      Device->WrFuse(Fuse);
      printf("done\n"); fflush(stdout);
    }
  }
  /*----------------- read lock bits ---------------------------------*/
  if(ReadLock != 0) {
    if(Device->RdLock == NULL) {
      fprintf(stderr,"*** no function for reading lock bits defined\n");
    } else {
      printf("reading lock bits.."); fflush(stdout);
      Device->RdLock(Lock);
      printf("done\n");
      printf("Lockbits:      ");
      if(Lock[0]==NOLOCK) { 
        printf("<none>\n");
      } else {
        for(i=0; Lock[i]!=NOLOCK;i++) { 
          printf("%s%c", GetLockBitId(Lock[i]), Lock[i+1]!=NOLOCK ? ',' : '\n');
        }
      } /* for(i=0; Lock[i]!=NOLOCK;i++) */
    }
  }
  /*----------------- read fuse bits ---------------------------------*/
  if(ReadFuse != 0) {
    if(Device->RdFuse == NULL) {
      fprintf(stderr,"*** no function for reading fuse bits defined\n");
    } else {
      printf("reading fuse bits.."); fflush(stdout);
      Device->RdFuse(Fuse);
      printf("done\n");
      printf("Fusebits:      ");
      if(Fuse[0]==NOFUSE) {
        printf("<none>\n");
      } else {
        for(i=0; Fuse[i]!=NOFUSE;i++) { 
          printf("%s%c", GetFuseBitId(Fuse[i]), Fuse[i+1]!=NOFUSE ? ',' : '\n');
        }
      } /* for(i=0; Fuse[i]!=NOFUSE;i++) */
    }
  }
  /*----------------- the end ----------------------------------------*/
  printf("Release STK200-Adaptor...\n");
  fflush(stdout);
  if((ErrCode = ReleaseStk200()) != 0) { 
    DBG_PRINTF(DBG_IO, "\t-> ErrorCode %ld\n", ErrCode);
    exit(2);
  }
  return(0);
}
